/*******************************************************************************
*    proxy 1.0.                                                                *
*    Copyright (C) 2001 Luca Bertoncello                                       *
*    Waltherstrae, 11D - 01067 Dresden Deutschland                            *
*    E-Mail: lucabert@lucabert.com, lucabert@lucabert.de                       *
*    http://www.lucabert.com/  http://www.lucabert.de/                         *
*                                                                              *
*    This program is free software; you can redistribute it and/or modify      *
*    it under the terms of the GNU General Public License as published by      *
*    the Free Software Foundation; version 2 of the License.                   *
*                                                                              *
*    This program is distributed in the hope that it will be useful,           *
*    but WITHOUT ANY WARRANTY; without even the implied warranty of            *
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
*    GNU General Public License for more details.                              *
*                                                                              *
*    You should have received a copy of the GNU General Public License         *
*    along with this program; if not, write to the Free Software               *
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#define SOCKET int
#define closesocket close
#define closesocket close
#define INVALID_SOCKET  (SOCKET)(~0)
#define bzero(A,B) (memset(A, 0, B))

/*
This program allow a PC in an external network (eg. Internet) to see
a specific port in a Server in an internal network (eg. Intranet).
Classic use of this kind of program is when there is a server with a
private IP, and a user from Internet have to launch a Web program in
this server.
*/


/*
Verbosity of the log.
This program logs through Syslog (printk)
*/

#define VERB_NO     0
#define VERB_MIN    1
#define VERB_MAX    2
#define VERBOSITY   VERB_MAX

/*
Core of the program.
It takes data from a socket and trasmit
them to the other.
*/
void proxy(SOCKET extSock, SOCKET intSock)
{
  int l, ok = 1, n;
  char s[1000];
  fd_set intFs, extFs, errIntFs, errExtFs;
  SOCKET nSock;

  if(extSock > intSock)
    nSock = extSock + 1;
  else
    nSock = intSock + 1;

// Set the sockets as Non Blocking
  fcntl(intSock, F_SETFL, O_NONBLOCK);
  fcntl(extSock, F_SETFL, O_NONBLOCK);

  while(ok)
    {
      n = recv(intSock, s, 100, 0);
      if(n > 0)
        send(extSock, s, n, 0);
      else
        if((errno != EAGAIN) || (n == 0))
          ok = 0;   // Socket closed
      n = recv(extSock, s, 100, 0);
      if(n > 0)
        send(intSock, s, n, 0);
      else
        if((errno != EAGAIN) || (n == 0))
          ok = 0;   // Socket closed
    }
}

/*
Connect to internal server, using
given IP and Port.
*/
SOCKET connectSock(char *ip, int port)
{
  struct sockaddr_in serv_addr;
  struct hostent *dati;
  SOCKET sockfd;

  bzero((char*)&serv_addr, sizeof(serv_addr));
  serv_addr.sin_family = AF_INET;
  serv_addr.sin_port = (u_short) htons((u_short) port);
  if((serv_addr.sin_addr.s_addr = inet_addr(ip)) == INADDR_NONE )
    {
      if((dati = (struct hostent*) gethostbyname(ip)) == NULL)
        {
          syslog(LOG_ERR, "Can't resolve '%s'\r\r\n", ip);
          return(1);
        }
      serv_addr.sin_addr.s_addr = *(long*)dati->h_addr;
    }

  if((sockfd = socket(AF_INET, SOCK_STREAM, 0)) != INVALID_SOCKET)
    {
      if(connect(sockfd, (struct sockaddr*)&serv_addr, sizeof(serv_addr)) >= 0)
        return sockfd;
      return 2;
    }

  return 1;
}

/*
Close an opened socket
*/
void closeSock(SOCKET sockfd)
{
  closesocket(sockfd);
}

/*
Open a communication with internal server
and run the Proxy function on it.
*/
void run(SOCKET extSock, char *intIP, int intPort)
{
  char str[1000];
  SOCKET intSock;

  intSock = connectSock(intIP, intPort);
  proxy(extSock, intSock);
  closeSock(intSock);
}

/*
Clean unused processes.
*/
void clean(int sig)
{
  int status;

  while(wait3(&status,WNOHANG,(struct rusage *) 0) >= 0)
    ;
}

/*
Main procedure.
It creates a server, listening on extPort.
When arrive a client, it lauchs run function,
to open a connection with internal server.
*/
int server(int extPort, char *intIP, int intPort)
{
  SOCKET sockfd, sock;
  struct sockaddr_in serv_addr, cli_addr;
  int clilen;
  pid_t pid;
  u_char *extIp;

  sockfd = socket(AF_INET, SOCK_STREAM, 0);
  bzero((char*)&serv_addr, sizeof(serv_addr));
  serv_addr.sin_family = AF_INET;
  serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  serv_addr.sin_port = (u_short) htons((u_short) extPort);
  if(bind(sockfd,(struct sockaddr*)&serv_addr,sizeof(serv_addr)) >= 0)
    {
      (void) signal(SIGCHLD, clean);
#if VERBOSITY != VERB_NO
      syslog(LOG_INFO, "Server active. Waiting for clients...");
#endif
      while(1)
        {
          listen(sockfd, 5);
          clilen = sizeof(cli_addr);
          sock = accept(sockfd, (struct sockaddr *)&cli_addr, &clilen );

          extIp = (u_char *) &cli_addr.sin_addr;
#if VERBOSITY != VERB_NO
          syslog(LOG_INFO, "Connection from %d.%d.%d.%d\n", extIp[0], extIp[1], extIp[2], extIp[3]);
#endif

          pid = fork();
          if(pid == -1)
            {
// CRITICAL ERROR! Hope not to happen!
              syslog(LOG_CRIT, "Out of fork table! EXIT!");
              closesocket(sockfd);
              closesocket(sock);
              exit(-1);
            }
          else
            if(pid == 0)
              {	/* Child process */
                run(sock, intIP, intPort);
#if VERBOSITY == VERB_MAX
          syslog(LOG_INFO, "Connection with %d.%d.%d.%d closed\n", extIp[0], extIp[1], extIp[2], extIp[3]);
#endif
                exit(0);
              }
          closesocket(sock);
        }

      closesocket(sockfd);
      return 0;
    }
  printf("Bind Error\n");
  return 2;
}

int main(int argc, char *argv[])
{
  if(argc < 4)
    {
      printf("Use:\n%s <external port> <internal IP> <internal port>\n", argv[0]);
      return -1;
    }
  else
    return server(atoi(argv[1]), argv[2], atoi(argv[3]));
}
